@@ -1,4 +1,4 @@
-// Copyright 2018 - 2021 Martin Dosch.
+// Copyright Martin Dosch.
 // Use of this source code is governed by the BSD-2-clause
 // license that can be found in the LICENSE file.
 
@@ -7,12 +7,12 @@ package main
 import (
 	"bufio"
 	"crypto/tls"
+	"errors"
 	"fmt"
 	"io"
 	"log"
 	"net"
 	"os"
-	"regexp"
 	"strings"
 	"time"
 
@@ -39,19 +39,17 @@ func readMessage(messageFilePath string) (string, error) {
 	// Check that message file is existing.
 	_, err = os.Stat(messageFilePath)
 	if os.IsNotExist(err) {
-		return output, err
+		return output, fmt.Errorf("readMessage: %w", err)
 	}
 
 	// Open message file.
 	file, err := os.Open(messageFilePath)
-	if err != nil {
-		return output, err
+	if errors.Unwrap(err) != nil {
+		return output, fmt.Errorf("readMessage: %w", err)
 	}
 	scanner := bufio.NewScanner(file)
 	scanner.Split(bufio.ScanLines)
-
 	for scanner.Scan() {
-
 		if output == "" {
 			output = scanner.Text()
 		} else {
@@ -59,15 +57,18 @@ func readMessage(messageFilePath string) (string, error) {
 		}
 	}
 
-	if err := scanner.Err(); err != nil {
-		if err != io.EOF {
-			return "", err
+	if err := scanner.Err(); errors.Unwrap(err) != nil {
+		if errors.Unwrap(err) != io.EOF {
+			return "", fmt.Errorf("readMessage: %w", err)
 		}
 	}
 
-	file.Close()
+	err = file.Close()
+	if errors.Unwrap(err) != nil {
+		fmt.Println("error while closing file:", err)
+	}
 
-	return output, err
+	return output, fmt.Errorf("readMessage: %w", err)
 }
 
 func main() {
@@ -85,7 +86,7 @@ func main() {
 
 	// Define command line flags.
 	flagHelp := getopt.BoolLong("help", 0, "Show help.")
-	flagHTTPUpload := getopt.StringLong("http-upload", 0, "", "Send a file via http-upload.")
+	flagHTTPUpload := getopt.StringLong("http-upload", 'h', "", "Send a file via http-upload.")
 	flagDebug := getopt.BoolLong("debug", 'd', "Show debugging info.")
 	flagServer := getopt.StringLong("jserver", 'j', "", "XMPP server address.")
 	flagUser := getopt.StringLong("username", 'u', "", "Username for XMPP account.")
@@ -104,8 +105,8 @@ func main() {
 		"Skip verification of TLS certificates (not recommended).")
 	flagRaw := getopt.BoolLong("raw", 0, "Send raw XML.")
 	flagListen := getopt.BoolLong("listen", 'l', "Listen for messages and print them to stdout.")
-	flagTimeout := getopt.IntLong("timeout", 0, 10, "Connection timeout in seconds.")
-	flagTLSMinVersion := getopt.IntLong("tls-version", 0, 12,
+	flagTimeout := getopt.IntLong("timeout", 0, defaultTimeout, "Connection timeout in seconds.")
+	flagTLSMinVersion := getopt.IntLong("tls-version", 0, defaultTLSMinVersion,
 		"Minimal TLS version. 10 (TLSv1.0), 11 (TLSv1.1), 12 (TLSv1.2) or 13 (TLSv1.3).")
 	flagVersion := getopt.BoolLong("version", 0, "Show version information.")
 	flagMUCPassword := getopt.StringLong("muc-password", 0, "", "Password for password protected MUCs.")
@@ -121,6 +122,8 @@ func main() {
 	flagOxImportPrivKey := getopt.StringLong("ox-import-privkey", 0, "",
 		"Import an existing private OpenPGP key.")
 	flagOxDeleteNodes := getopt.BoolLong("ox-delete-nodes", 0, "Delete existing OpenPGP nodes on the server.")
+	flagOOBFile := getopt.StringLong("oob-file", 0, "", "URL to send a file as out of band data.")
+	flagHeadline := getopt.BoolLong("headline", 0, "Send message as type headline.")
 
 	// Parse command line flags.
 	getopt.Parse()
@@ -128,7 +131,7 @@ func main() {
 	switch {
 	case *flagHelp:
 		// If requested, show help and quit.
-		getopt.Usage()
+		getopt.PrintUsage(os.Stdout)
 		os.Exit(0)
 	case *flagVersion:
 		// If requested, show version and quit.
@@ -147,6 +150,10 @@ func main() {
 	case *flagHTTPUpload != "" && *flagMessageFile != "":
 		log.Fatal("You can't send a message while using" +
 			" http upload.")
+	case *flagOx && *flagOOBFile != "":
+		log.Fatal("No encryption possible for OOB data.")
+	case *flagHeadline && *flagChatroom:
+		log.Fatal("Can't use message type headline for groupchat messages.")
 	}
 
 	// Read recipients from command line and quit if none are specified.
@@ -163,7 +170,7 @@ func main() {
 	if *flagUser == "" || *flagPassword == "" {
 		// Read configuration from file.
 		config, err := parseConfig(*flagFile)
-		if err != nil {
+		if errors.Unwrap(err) != nil {
 			log.Fatal("Error parsing ", *flagFile, ": ", err)
 		}
 		// Set connection options according to config.
@@ -219,7 +226,7 @@ func main() {
 	}
 
 	// Timeout
-	timeout := time.Duration(*flagTimeout * 1000000000)
+	timeout := time.Duration(*flagTimeout) * time.Second
 
 	// Use ALPN
 	var tlsConfig tls.Config
@@ -227,13 +234,13 @@ func main() {
 	tlsConfig.NextProtos = append(tlsConfig.NextProtos, "xmpp-client")
 	tlsConfig.InsecureSkipVerify = *flagSkipVerify
 	switch *flagTLSMinVersion {
-	case 10:
+	case defaultTLS10:
 		tlsConfig.MinVersion = tls.VersionTLS10
-	case 11:
+	case defaultTLS11:
 		tlsConfig.MinVersion = tls.VersionTLS11
-	case 12:
+	case defaultTLS12:
 		tlsConfig.MinVersion = tls.VersionTLS12
-	case 13:
+	case defaultTLS13:
 		tlsConfig.MinVersion = tls.VersionTLS13
 	default:
 		fmt.Println("Unknown TLS version.")
@@ -262,52 +269,50 @@ func main() {
 	// Read message from file.
 	if *flagMessageFile != "" {
 		message, err = readMessage(*flagMessageFile)
-		if err != nil {
+		if errors.Unwrap(err) != nil {
 			log.Fatal(err)
 		}
 	}
 
 	// Skip reading message if '-i' or '--interactive' is set to work with e.g. 'tail -f'.
-	// Also for listening mode.
-	if !*flagInteractive && !*flagListen && *flagHTTPUpload == "" {
-		if message == "" {
-
-			scanner := bufio.NewScanner(os.Stdin)
-			for scanner.Scan() {
-
-				if message == "" {
-					message = scanner.Text()
-				} else {
-					message = message + "\n" + scanner.Text()
-				}
+	// Also for listening mode and Ox key handling.
+	if !*flagInteractive && !*flagListen && *flagHTTPUpload == "" &&
+		!*flagOxDeleteNodes && *flagOxImportPrivKey == "" &&
+		!*flagOxGenPrivKeyX25519 && !*flagOxGenPrivKeyRSA && *flagOOBFile == "" &&
+		message == "" {
+		scanner := bufio.NewScanner(os.Stdin)
+		for scanner.Scan() {
+			if message == "" {
+				message = scanner.Text()
+			} else {
+				message = message + "\n" + scanner.Text()
 			}
+		}
 
-			if err := scanner.Err(); err != nil {
-				if err != io.EOF {
-					log.Fatal(err)
-				}
+		if err := scanner.Err(); err != nil {
+			if errors.Unwrap(err) != io.EOF {
+				log.Fatal(err)
 			}
 		}
 	}
 
 	// Remove invalid code points.
-	message = strings.ToValidUTF8(message, "")
-	reg := regexp.MustCompile(`[\x{0000}-\x{0008}\x{000B}\x{000C}\x{000E}-\x{001F}]`)
-	message = reg.ReplaceAllString(message, "")
-
+	message = validUTF8(message)
 	// Exit if message is empty.
-	if message == "" && !*flagInteractive && !*flagListen && *flagHTTPUpload == "" {
+	if message == "" && !*flagInteractive && !*flagListen && !*flagOxGenPrivKeyRSA &&
+		!*flagOxGenPrivKeyX25519 && *flagOxImportPrivKey == "" &&
+		!*flagOxDeleteNodes && *flagHTTPUpload == "" && *flagOOBFile == "" {
 		os.Exit(0)
 	}
 
 	// Connect to server.
 	client, err := connect(options, *flagDirectTLS)
-	if err != nil {
+	if errors.Unwrap(err) != nil {
 		log.Fatal(err)
 	}
 
-	iqc := make(chan xmpp.IQ, 100)
-	msgc := make(chan xmpp.Chat, 100)
+	iqc := make(chan xmpp.IQ, defaultBufferSize)
+	msgc := make(chan xmpp.Chat, defaultBufferSize)
 	go rcvStanzas(client, iqc, msgc)
 
 	for _, r := range getopt.Args() {
@@ -315,7 +320,7 @@ func main() {
 		re.Jid = r
 		if *flagOx {
 			re.OxKeyRing, err = oxGetPublicKeyRing(client, iqc, r)
-			if err != nil {
+			if errors.Unwrap(err) != nil {
 				re.OxKeyRing = nil
 				fmt.Println("ox: error fetching key for", r+":", err)
 			}
@@ -326,7 +331,7 @@ func main() {
 	// Check that all recipient JIDs are valid.
 	for i, recipient := range recipients {
 		validatedJid, err := MarshalJID(recipient.Jid)
-		if err != nil {
+		if errors.Unwrap(err) != nil {
 			log.Fatal(err)
 		}
 		recipients[i].Jid = validatedJid
@@ -335,52 +340,52 @@ func main() {
 	switch {
 	case *flagOxGenPrivKeyX25519:
 		validatedOwnJid, err := MarshalJID(user)
-		if err != nil {
+		if errors.Unwrap(err) != nil {
 			log.Fatal(err)
 		}
 		err = oxGenPrivKey(validatedOwnJid, client, iqc, *flagOxPassphrase, "x25519")
-		if err != nil {
+		if errors.Unwrap(err) != nil {
 			log.Fatal(err)
 		}
 		os.Exit(0)
 	case *flagOxGenPrivKeyRSA:
 		validatedOwnJid, err := MarshalJID(user)
-		if err != nil {
+		if errors.Unwrap(err) != nil {
 			log.Fatal(err)
 		}
 		err = oxGenPrivKey(validatedOwnJid, client, iqc, *flagOxPassphrase, "rsa")
-		if err != nil {
+		if errors.Unwrap(err) != nil {
 			log.Fatal(err)
 		}
 		os.Exit(0)
 	case *flagOxImportPrivKey != "":
 		validatedOwnJid, err := MarshalJID(user)
-		if err != nil {
+		if errors.Unwrap(err) != nil {
 			log.Fatal(err)
 		}
 		err = oxImportPrivKey(validatedOwnJid, *flagOxImportPrivKey,
 			client, iqc)
-		if err != nil {
+		if errors.Unwrap(err) != nil {
 			log.Fatal(err)
 		}
 		os.Exit(0)
 	case *flagOxDeleteNodes:
 		validatedOwnJid, err := MarshalJID(user)
-		if err != nil {
+		if errors.Unwrap(err) != nil {
 			log.Fatal(err)
 		}
 		err = oxDeleteNodes(validatedOwnJid, client, iqc)
-		if err != nil {
+		if errors.Unwrap(err) != nil {
 			log.Fatal(err)
 		}
 		os.Exit(0)
 	case *flagOx:
 		validatedOwnJid, err := MarshalJID(user)
-		if err != nil {
+		if errors.Unwrap(err) != nil {
 			log.Fatal(err)
 		}
 		oxPrivKey, err = oxGetPrivKey(validatedOwnJid, *flagOxPassphrase)
-		if err != nil {
+		if errors.Unwrap(err) != nil {
 			log.Fatal(err)
 		}
 	}
@@ -390,10 +395,25 @@ func main() {
 			*flagHTTPUpload)
 	}
 
+	if *flagOOBFile != "" {
+		// Remove invalid UTF8 code points.
+		message = validUTF8(*flagOOBFile)
+		// Check if the URI is valid.
+		uri, err := validURI(message)
+		if errors.Unwrap(err) != nil {
+			log.Fatal(err)
+		}
+		message = uri.String()
+	}
+
 	var msgType string
-	msgType = "chat"
+	if *flagHeadline {
+		msgType = strHeadline
+	} else {
+		msgType = strChat
+	}
 	if *flagChatroom {
-		msgType = "groupchat"
+		msgType = strGroupchat
 		// Join the MUCs.
 		for _, recipient := range recipients {
 			if *flagMUCPassword != "" {
@@ -403,7 +423,7 @@ func main() {
 			} else {
 				_, err = client.JoinMUCNoHistory(recipient.Jid, alias)
 			}
-			if err != nil {
+			if errors.Unwrap(err) != nil {
 				log.Fatal(err)
 			}
 		}
@@ -415,22 +435,21 @@ func main() {
 		}
 		// Send raw XML
 		_, err = client.SendOrg(message + "\n")
-		if err != nil {
+		if errors.Unwrap(err) != nil {
 			log.Fatal(err)
 		}
 	case *flagInteractive:
 		// Send in endless loop (for usage with e.g. "tail -f").
-		var reader = bufio.NewReader(os.Stdin)
+		reader := bufio.NewReader(os.Stdin)
 		for {
 			message, err = reader.ReadString('\n')
-			if err != nil {
+			message = strings.TrimSuffix(message, "\n")
+			if errors.Unwrap(err) != nil {
 				log.Fatal("failed to read from stdin")
 			}
 
 			// Remove invalid code points.
-			message = strings.ToValidUTF8(message, "")
-			reg := regexp.MustCompile(`[\x{0000}-\x{0008}\x{000B}\x{000C}\x{000E}-\x{001F}]`)
-			message = reg.ReplaceAllString(message, "")
+			message = validUTF8(message)
 			if message == "" {
 				continue
 			}
@@ -440,21 +459,23 @@ func main() {
 					if recipient.OxKeyRing == nil {
 						continue
 					}
-					oxMessage, err := oxEncrypt(client, iqc, oxPrivKey,
+					oxMessage, err := oxEncrypt(client, oxPrivKey,
 						recipient.Jid, recipient.OxKeyRing, message)
-					if err != nil {
+					if errors.Unwrap(err) != nil {
 						fmt.Println("Ox: couldn't encrypt to",
 							recipient.Jid)
 						continue
 					}
 					_, err = client.SendOrg(oxMessage + "\n")
-					if err != nil {
+					if errors.Unwrap(err) != nil {
 						log.Fatal(err)
 					}
 				default:
-					_, err = client.Send(xmpp.Chat{Remote: recipient.Jid,
-						Type: msgType, Text: message})
-					if err != nil {
+					_, err = client.Send(xmpp.Chat{
+						Remote: recipient.Jid,
+						Type:   msgType, Text: message,
+					})
+					if errors.Unwrap(err) != nil {
 						log.Fatal(err)
 					}
 				}
@@ -467,7 +488,7 @@ func main() {
 			switch {
 			case isOxMsg(v) && *flagOx:
 				msg, t, err := oxDecrypt(v, client, iqc, user, oxPrivKey)
-				if err != nil {
+				if errors.Unwrap(err) != nil {
 					log.Println(err)
 					continue
 				}
@@ -476,9 +497,9 @@ func main() {
 				}
 				var bareFrom string
 				switch v.Type {
-				case "chat":
+				case strChat:
 					bareFrom = strings.Split(v.Remote, "/")[0]
-				case "groupchat":
+				case strGroupchat:
 					bareFrom = v.Remote
 				default:
 					bareFrom = strings.Split(v.Remote, "/")[0]
@@ -508,9 +529,9 @@ func main() {
 				}
 				var bareFrom string
 				switch v.Type {
-				case "chat":
+				case strChat:
 					bareFrom = strings.Split(v.Remote, "/")[0]
-				case "groupchat":
+				case strGroupchat:
 					bareFrom = v.Remote
 				default:
 					bareFrom = strings.Split(v.Remote, "/")[0]
@@ -536,9 +557,21 @@ func main() {
 			}
 			switch {
 			case *flagHTTPUpload != "":
-				_, err = client.Send(xmpp.Chat{Remote: recipient.Jid,
-					Type: msgType, Ooburl: message, Text: message})
-				if err != nil {
+				_, err = client.Send(xmpp.Chat{
+					Remote: recipient.Jid,
+					Type:   msgType, Ooburl: message, Text: message,
+				})
+				if errors.Unwrap(err) != nil {
+					fmt.Println("Couldn't send message to",
+						recipient.Jid)
+				}
+			// (Hopefully) temporary workaround due to go-xmpp choking on URL encoding.
+			// Once this is fixed in the lib the http-upload case above can be reused.
+			case *flagOOBFile != "":
+				_, err = client.SendOrg("<message to='" + recipient.Jid + "' type='" +
+					msgType + "'><body>" + message + "</body><x xmlns='jabber:x:oob'><url>" +
+					message + "</url></x></message>")
+				if errors.Unwrap(err) != nil {
 					fmt.Println("Couldn't send message to",
 						recipient.Jid)
 				}
@@ -546,20 +579,22 @@ func main() {
 				if recipient.OxKeyRing == nil {
 					continue
 				}
-				oxMessage, err := oxEncrypt(client, iqc, oxPrivKey,
+				oxMessage, err := oxEncrypt(client, oxPrivKey,
 					recipient.Jid, recipient.OxKeyRing, message)
-				if err != nil {
+				if errors.Unwrap(err) != nil {
 					fmt.Println("Ox: couldn't encrypt to", recipient.Jid)
 					continue
 				}
 				_, err = client.SendOrg(oxMessage + "\n")
-				if err != nil {
+				if errors.Unwrap(err) != nil {
 					log.Fatal(err)
 				}
 			default:
-				_, err = client.Send(xmpp.Chat{Remote: recipient.Jid,
-					Type: msgType, Text: message})
-				if err != nil {
+				_, err = client.Send(xmpp.Chat{
+					Remote: recipient.Jid,
+					Type:   msgType, Text: message,
+				})
+				if errors.Unwrap(err) != nil {
 					log.Fatal(err)
 				}
 			}
@@ -568,5 +603,5 @@ func main() {
 
 	// Wait for a short time as some messages are not delievered by the server
 	// if the connection is closed immediately after sending a message.
-	time.Sleep(100 * time.Millisecond)
+	time.Sleep(defaultSleepTime * time.Millisecond)
 }
